{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleInstances, MultiParamTypeClasses, OverlappingInstances #-}

module F4L.Actions (
    module F4L.F4LMonad
  , addFiles
  , addTestFiles
  , autoBackup
  , backupPath
  , coverage
  , dumpOps
  , dumpFacts
  , extraSession
  , findStorage
  , getDBFile
  , inSession
  , f4lState
  , findLocalFileCandidates
  , forceGetEnvId
  , fromDB
  , importExportSessions
  , initDB
  , labelStorage
  , labelEnv
  , labelStorage
  , listStorageContents
  , loadCached
  , loadOps
--  , readOps, sliceSessions, readFullOps
  , onErrorDie
  , onErrorUseDefault
  , onErrorWarn
  , onErrorWarnWithDefault
  , restore
  , saveCache
  , sessions
  , storages
  , stats
  , withKnowledge
  ) where

import F4L
import F4L.BufferMonad
import F4L.SeenTree as ST hiding ( removeFile )
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import qualified F4L.FSMonad
import F4L.Knowledge
import F4L.Op
import F4L.Utils hiding( fileSize )

import qualified Codec.Compression.GZip as GZ
import Control.DeepSeq
-- import Control.Concurrent( threadDelay )
import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..), throwError )
import Control.Monad.Reader
-- import qualified Control.Monad.State as SL
import Control.Monad.State.Strict
import Data.Array.IArray( elems )
import Data.Binary( encode )
import qualified Data.Binary as B
import Data.Binary.Get
import Data.Binary.Put( runPut )
-- import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BL
import qualified Data.ByteString.Short as SBS
import Data.Either( partitionEithers )
import Data.Function( on )
import Data.List( foldl', sort, sortBy, inits, isPrefixOf, unfoldr )
-- import Data.Map( Map )
import qualified Data.Map as M -- TODO: use strict ?
import qualified Data.Map.Strict as SM
-- import Data.Maybe( maybeToList )
import Data.Monoid( mappend )
-- import Data.Set( Set )
import qualified Data.Set as S
-- import qualified Data.Text as T
import System.FilePath( combine, dropTrailingPathSeparator, takeDirectory, joinPath, splitDirectories )


{---------------------------------------- Utils ----------------------------------------}

class VerboseMonad m where
  say :: VerbosityLevel -> String -> m ()
  outputLineMsg :: String -> m ()
  cleanLine :: m ()

instance FSMonad m => VerboseMonad ( GenF4LMonad k s m ) where
  say l msg = do
    v <- use st_verbosity
    if l <= verb_level v then outputMsg msg else return ()

  outputLineMsg msg =
    whenM ( use $ st_verbosity . to verb_tracing ) $ F4L.FSMonad.outputLineMsg msg

  cleanLine =
    whenM ( use $ st_verbosity . to verb_tracing ) F4L.FSMonad.cleanLine

-- instance ( Monad m, VerboseMonad m ) => VerboseMonad ( ErrorT Error m ) where
instance ( MonadTrans t, Monad m, VerboseMonad m ) => VerboseMonad ( t m ) where
  say l msg = lift $ say l msg
  outputLineMsg = lift . outputLineMsg
  cleanLine = lift cleanLine


fileId :: ( FSMonad m, BufferMonad m ) => FilePath -> ErrorT Error ( F4LMonad k m ) FileId
-- fileId path = processLengthAndBLFromFile path $ \ l bs -> FileId l ( sha1 bs )
fileId path = actOnLengthAndBLFromFile path act
  where
    act l bs = do
      let s = sha1 bs
      let res = FileId l s
      s `deepseq` return res


checkedCopyFile :: DelFSMonad m => FileId -> FilePath -> FilePath -> ErrorT Error m ()
checkedCopyFile ( FileId l h ) source dest = do
  writeRes <- actOnLengthAndBLFromFile source checkedWrite
  case writeRes of
    Left err -> throwError err
    Right res -> return res
  where
    checkedWrite i bl = runErrorT $ do
      when ( i /= l ) $
        throwError $ "Length mismatch copying " ++ source ++ " to " ++ dest ++ ": should be " ++ show l ++ ", is " ++ show i
      ctx <- writeBLToFileAndDo dest newSha1 ( liftM ( liftM return ) updateSha1' ) bl
      let rh = toSha1 ctx
      when ( rh /= h ) $ do
        removeFile dest
        throwError $ "Hash mismatch copying " ++ source ++ " to " ++ dest ++ ": should be " ++ hashToS h ++ ", is " ++ hashToS rh
      return ()


checkedCopyFileCandidates :: DelFSMonad m => FileId -> [ FilePath ] -> FilePath -> String -> ErrorT Error ( F4LMonad k m ) ()
checkedCopyFileCandidates fid candidates destPath preMessage =
  foldl mplus ( throwError "No candidate for file" ) $ flip map candidates $ \ sourcePath -> do
    outputLineMsg $ preMessage ++ "Copying " ++ sourcePath ++ " to " ++ destPath
    say Chatty $ "Copying " ++ sourcePath ++ " to " ++ destPath
    checkedCopyFile fid sourcePath destPath


findLocalFileCandidates :: ( FSMonad m, BufferMonad m, KnownDataClass k ) 
                           => FileId -> F4LMonad k m [ FilePath ]
findLocalFileCandidates fid = do
  storages <- storagesForEnv
  fd <- use $ st_knowledge . kd_files . to ( SM.findWithDefault noFileData fid  )
  return $ fileCandidates storages fd


fileCandidates :: SM.Map StorageId [ FilePath ] -> FileData -> [ FilePath ]
fileCandidates storages dat =
  [ combine sseen path
  | ( sid, path ) <- map ( mapSnd shortToPath ) $ S.toList ( fd_seen dat )
  , sseen <- M.findWithDefault [] sid storages
  ]


{-
findFileWithSize :: ( FSMonad m, BufferMonad m ) => Int -> [ FilePath ] -> m ( Maybe FilePath )
findFileWithSize _ [] = return Nothing
findFileWithSize i ( x : xs ) = do
  e <- doesFileExist x
  case e of
    True -> do
      es <- runErrorT $ fileSize x
      case es of
        Right s | s == i -> return ( Just x )
        _ -> findFileWithSize i xs
    False ->
      findFileWithSize i xs
-}


sidPathCmp :: ( StorageId, FilePath ) -> ( StorageId, FilePath ) -> Ordering
sidPathCmp = ( compare `on` fst ) `mappend` ( compare `on` ( splitDirectories . snd ) )


withKnowledge :: Monad m => k -> GenF4LMonad k s m ()
withKnowledge = const $ return ()


onErrorDie :: Monad m => ErrorT Error m a -> m a
onErrorDie a = do
  res <- runErrorT a
  case res of
    Left err -> error err
    Right x -> return x


onErrorUseDefault :: ( VerboseMonad m, FSMonad m ) => a -> ErrorT Error m a -> m a
onErrorUseDefault def a = liftM ( either ( const def ) id ) $ runErrorT a


onErrorWarn :: ( VerboseMonad m, FSMonad m ) => ErrorT Error m () -> m ()
onErrorWarn = onErrorWarnWithDefault ()


onErrorWarnWithDefault :: ( VerboseMonad m, FSMonad m ) => a -> ErrorT Error m a -> m a
onErrorWarnWithDefault def a = do
  res <- runErrorT a
  case res of
    Left err -> say Quiet ( "Warning: " ++ err ) >> return def
    Right x -> return x


data RichPath = RichPath {
    rp_sid :: !StorageId -- ^ Storage ID
  , rp_sPath :: !FilePath -- ^ Storage path
  , rp_relPath :: !FilePath -- ^ Path wrt storage
  , rp_canPath :: !FilePath -- ^ Canonical path
  } deriving Show


{---------------------------------------- Actions ----------------------------------------}


interactivePrompt def msg = do
  i <- use st_interactive
  case i of
    Ask -> prompt msg
    UseDefault -> return def
    Use answer -> st_interactive .= UseDefault >> return answer


-- | Add component to RichPath
subRichPath :: FilePath -> RichPath -> RichPath
subRichPath fp rp = rp { rp_relPath = combine ( rp_relPath rp ) fp
                       , rp_canPath = combine ( rp_canPath rp ) fp
                       }

f4lState :: Knowledge k 
            => MetaDB -> Verbosity -> Interactivity -> Bool -> Bool -> s -> F4LState k s
f4lState a b c d e s = F4LState a b c d e initialOpEnv s M.empty 0 noKnowledge

initDB :: ( FSMonad m, BufferMonad m ) => ErrorT Error ( TopF4LMonad k m ) ()
initDB = lift getDBFile >>= flip appendToFile ""


extraSession :: ( FSMonad m, BufferMonad m ) => TopF4LMonad k m () -> GenF4LMonad k s m ()
extraSession act = do
  pre_s <- use st_session
  state <- get
  end_state <- lift $ execStateT act ( st_session .~ () $ state )
  put $ st_session .~ pre_s $ end_state

inSession :: ( FSMonad m, BufferMonad m, Knowledge k ) 
             => EnvironmentId -> F4LMonad k m a -> GenF4LMonad k s m a
inSession env act = do
  ts <- lift $ getNow
  sessid <- generateUuid
  inSession' env ts sessid act

inSession' :: ( FSMonad m, BufferMonad m, Knowledge k ) 
             => EnvironmentId -> TimeStamp -> SessionId -> F4LMonad k m a -> GenF4LMonad k s m a
inSession' env ts sessid act = do
  pre_s <- use st_session
  state <- get
  ( res, end_state ) <- lift $ runStateT ( execOp (StartSession undefined undefined env ts 0) >> act >>= \ res -> finalize env ts sessid >> return res ) ( st_session .~ ActiveSession env ts $ state )
  put $ st_session .~ pre_s $ end_state
  return res


finalize :: ( FSMonad m, BufferMonad m, Knowledge k ) 
            => EnvironmentId -> TimeStamp -> SessionId -> F4LMonad k m ()
finalize env ts sessid = do
  l <- bufferLength
  when ( l > 0 ) $ do
    db <- getDBFile
    addOp EndSession
    (buf, bufl) <- freezedBuffer
    onErrorDie $ appendBLsToFile db [ 
        encode ( StartSession currentVersion sessid env ts bufl )
      , BL.pack ( take bufl $ elems buf ) ]
  emptyBuffer


getDBFile :: Monad m => GenF4LMonad k s m FilePath
getDBFile = use ( conf_db . to mdb_file )


execOp :: ( Monad m, Knowledge k ) => Op -> GenF4LMonad k s m [Fact]
execOp op = do
  openv <- use st_opEnv
  let ( facts, newOpenv ) = runState ( applyOp op ) openv
  st_opEnv .= newOpenv
  flip mapM_ facts $ \ fact -> st_knowledge %= applyFact fact
  return facts


addOp :: ( BufferMonad m, Knowledge k ) => Op -> F4LMonad k m ()
addOp op = addOp' op >> return ()


addOp' :: ( BufferMonad m, Knowledge k ) => Op -> F4LMonad k m [Fact]
addOp' op = do
  addToBuffer $ encode op
  execOp op


everyN :: Monad m => Int -> F4LMonad k m () -> F4LMonad k m ()
everyN n act = do
  i <- use st_debugCounter
  if i >= n - 1 then
    st_debugCounter .= 0 >> act
    else st_debugCounter += 1


setPath :: ( BufferMonad m, Knowledge k ) => FilePath -> F4LMonad k m ()
setPath path = do
  curpath <- use $ st_opEnv . openv_path
  if curpath == path then
    return()
    else do
    mfid <- use $ st_opEnv . openv_fid
    case mfid of
         Just fid | path == uncurry combine ( fidToBDest fid ) -> do
           addOp BackupPath
         _ -> do
           let common_l = length $ takeWhile ( uncurry (==) ) $ zip curpath path
           let common_t = length $ takeWhile ( uncurry (==) ) $ zip ( reverse $ drop common_l curpath ) ( reverse $ drop common_l path )
           if ( common_t >= 2 ) then
             addOp $ PatchPath ( fromIntegral $ length curpath - common_l ) ( fromIntegral common_t ) $ drop common_l $ reverse $ drop common_t $ reverse path
             else addOp $ RetailPath ( fromIntegral $ length curpath - common_l ) $ drop common_l path


setFid :: ( BufferMonad m, Knowledge k ) => FileId -> F4LMonad k m ()
setFid fid = do
  opEnv <- use st_opEnv
  case fidShorthand fid opEnv of
    Just s -> addOp $ WithKnownFid s
    Nothing -> addOp $ WithFid fid


attrOp :: FilePathAttr a -> Op
attrOp a = case a of
  Require req -> WithPriority req
  Tag tag -> WithTag tag
  Property k v -> WithProperty k v


addTestFile :: ( BufferMonad m, Knowledge k ) => Int -> F4LMonad k m ()
addTestFile i = do
  let fid = testFid i
  setFid fid
  setPath $ "test-" ++ show i
  addOp SeenMark
  return ()


addTestFiles :: ( BufferMonad m, Knowledge k ) => Int -> F4LMonad k m ()
addTestFiles i = do
  withStorage "-" "/"
  mapM_ addTestFile [ 1 .. i ]
  return ()


forceGetEnvId :: ( FSMonad m ) => FilePath -> m EnvironmentId
forceGetEnvId dir = do
  f <- runErrorT $ readFromFile $ combine dir "env-id"
  case f of
    Right env -> return env
    Left _err -> do
      env <- generateUuid
      onErrorDie $ createDirectoryIfMissing False dir >> appendToFile ( combine dir "env-id" ) env
      return env


getStorageId :: ( FSMonad m, BufferMonad m ) => FilePath -> m ( Maybe StorageId )
getStorageId dir = do
  f <- runErrorT $ readFromFile $ combine dir storageIdFile
  return $ case f of
    Right sid -> Just sid
    Left _err -> Nothing


initStorageId :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                 => FilePath -> ErrorT Error ( F4LMonad k m ) StorageId
initStorageId path = do
  label <- lift $ do
    say Chatty $ "Init-ing storage on " ++ path
    env <- use st_envid
    let deflabel = path ++ " on " ++ env
    s <- interactivePrompt deflabel $ "Add label to created storage. Default is " ++ deflabel
    return $ if s == "" then deflabel else s
  sid <- lift generateUuid
  appendToFile ( combine path storageIdFile ) sid
  lift $ addOp $ AddStorageLabel sid label
  return sid


forceGetStorageId :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                     => FilePath -> ErrorT Error ( F4LMonad k m ) StorageId
forceGetStorageId path = do
  mpre <- lift $ getStorageId path
  case mpre of
    Just id -> return id
    Nothing -> initStorageId path


fakeGetStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                  => FilePath -> ErrorT Error ( F4LMonad k m ) RichPath
fakeGetStorage path = do
  canPath <- lift $ canonicalizePath path
  return $ RichPath "" canPath "" canPath


forceGetStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                   => FilePath -> ErrorT Error ( F4LMonad k m ) RichPath
forceGetStorage path = do
  say Chatty $ "Forcing storage on " ++ path
  mres <- lift $ getStorage path
  case mres of
    Just r -> say Timid ( "Found " ++ show r ) >> return r
    Nothing -> do
      canDir <- lift $ canonicalizePath path
      let paths = map joinPath $ drop 1 $ inits $ splitDirectories canDir
      let pathchoices = zip [1..] $ reverse paths
      dpath <- do
        s <- lift $ interactivePrompt "1" $ unlines $
             [ "No storage for " ++ path
             , "Create one" ] ++
             map ( \ (i, v) -> show i ++ " - " ++ v ) pathchoices
        return $ snd $ head $ filter ( ( s == ) . show . fst ) pathchoices
      _sid <- initStorageId dpath
      lift $ st_storages %= M.delete dpath
      forceGetStorage path


-- | returns storage, sPath, relPath
--   performs path canonicalization for sPath
getStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) => FilePath -> F4LMonad k m ( Maybe RichPath )
getStorage path = do
  smap <- use st_storages
  canPath <- canonicalizePath path
  let parts = splitDirectories canPath
  let dirs = reverse $ zipWith splitAt [ 1 .. length parts ] $ repeat parts
  msidpath <- testDirs smap dirs
  return $ liftM ( \ ( sid, sPath, relPath ) -> RichPath sid sPath relPath canPath ) msidpath
  where
    test smap dir = do
      case M.lookup dir smap of
        Just x @ ( Just _sid ) -> return x
        Just Nothing -> return Nothing
        Nothing -> do
          msid <- getStorageId dir
          st_storages %= M.insert dir msid
          return msid
    testDirs _ [] = return Nothing
    testDirs smap ( ( dparts, fparts ) : xs ) = do
      let dir = joinPath dparts
      res <- test smap dir
      case res of
        Nothing -> testDirs smap xs
        Just sid -> return $ Just ( sid, dir, joinPath fparts )


withStorage :: ( BufferMonad m, Knowledge k ) => StorageId -> FilePath -> F4LMonad k m ()
withStorage sid path = do
  oldS <- use $ st_opEnv . openv_storage
  when ( oldS /= Just ( sid, path ) ) $ addOp $ WithStorage sid $ dropTrailingPathSeparator path


-- | Ops are only added on no error
addFile :: ( FSMonad m, BufferMonad m, Knowledge k, KnownDataClass k ) 
           => RichPath -> Bool -> FileId
           -> ErrorT Error ( F4LMonad k m ) ()
addFile rp samePosition fid = do
  let ( sid, sPath, relPath, path ) = ( rp_sid rp, rp_sPath rp, rp_relPath rp, rp_canPath rp )
  say Chatty $ "Adding " ++ sid ++ ":" ++ sPath ++ ":" ++ relPath
  withExif <- lift $ use st_withExif
  exif_props <- if withExif && not samePosition then lift ( exifProperties path ) else return []
  lift $ do
    withStorage sid sPath
    when ( not samePosition ) $ do
      setFid fid
      setPath relPath
      -- TODO: only add new properties
      mapM_ (addOp . attrOp) exif_props
      addOp SeenMark


{-
debugState :: ( FSMonad m, MonadIO m, Show k ) => F4LState k -> F4LMonad k m ()
debugState st = do
  say Quiet $ " " ++ ( show $ length $ show st )
  liftIO $ threadDelay 100000
  return ()
-}


addFiles :: ( FSMonad m, BufferMonad m, Knowledge k, SeenStoragesClass k, KnownDataClass k, Show k ) => Maybe Priority -> [FilePath] -> F4LMonad k m ()
addFiles mpri paths = do
  ( withError, richPaths ) <- liftM partitionEithers $ flip mapM paths $ \ p -> do
    res <- runErrorT $ forceGetStorage p
    case res of
      Right res -> return $ Right res
      Left err -> return $ Left ( p, err )
  flip mapM_ withError $ \ ( p, err ) -> do
    say Quiet $ "Warning: error setting StorageId for " ++ p ++ ": " ++ err
  let sidLocations = SM.fromList [ ( rp_sid x, rp_sPath x ) | x <- richPaths ]
  say SayAll $ "richPaths: " ++ show richPaths
  let rootSidPaths = dominants sidPathCmp dominates $ map toSidPath richPaths
  say SayAll $ "rootSidPaths: " ++ show rootSidPaths
  withRescan <- use st_withRescan
  -- accumulate files seen in same position with some size
  say Chatty "Adding files"
  alreadySeen <- use $ st_knowledge . kd_contents . to ( SM.map ( \ a -> (a, False) ) . seenFiles rootSidPaths )
  missing <- liftM ( SM.keys . SM.filter ( not . snd ) ) $ execStateT ( mapM_ ( addFile' True withRescan ) richPaths ) alreadySeen
  mapM_ ( addMissing sidLocations ) $ sort missing
  cleanLine
  
  where
    toSidPath :: RichPath -> ( StorageId, FilePath )
    toSidPath rp = ( rp_sid rp, rp_relPath rp )
    seenFiles ::  [(StorageId, FilePath)] -> SM.Map StorageId SeenTree -> SM.Map (StorageId, FilePath) FileId
    seenFiles roots strees = SM.fromList $ concat $ [ seenFiles' ( SM.findWithDefault ST.empty sid strees ) sid fp | ( sid, fp ) <- roots ]
    seenFiles' stree sid fp = [ ((sid, joinPath $ fp : pcomp ), fid) | (pcomp, fid) <- subFiles ( splitDirectories fp ) stree ]
    addMissing :: ( BufferMonad m, Monad m, Knowledge k, KnownDataClass k ) 
                  => SM.Map StorageId FilePath -> ( StorageId, FilePath ) -> F4LMonad k m  ()
    addMissing sm ( sid, fp ) = case SM.lookup sid sm of
      Nothing -> error $ "Aborted: unable to find storage " ++ sid ++ " within " ++ show sm
      Just sp -> do
        withStorage sid sp
        setPath fp
        addOp Missing

    addFile' :: ( FSMonad m, BufferMonad m, Knowledge k, Show k, KnownDataClass k ) 
                => Bool -> Bool -> RichPath -> StateT ( SM.Map (StorageId, FilePath) ( FileId, Bool ) ) ( F4LMonad k m ) ()
    addFile' withSetPri withRescan rp = do
      isDir <- doesDirectoryExist $ rp_canPath rp
      -- everyN 1000 $ get >>= say Quiet . show . length . show
      if isDir then do
        say Timid $ "Descending into " ++ show ( rp_canPath rp )
        subfps <- lift $ onErrorWarnWithDefault [] $ getDirectoryContents ( rp_canPath rp ) >>= return . filter ( not . flip elem [ ".", ".." ] ) >>= return . sort
        mapM_ ( \ fp -> addFile' False withRescan ( subRichPath fp rp ) ) subfps
        else do
        -- On error report size -1 to make sure no collision happens
        -- TODO: choose better way?
        s <- lift $ onErrorWarnWithDefault (-1) $ fileSize $ rp_canPath rp
        let k = ( rp_sid rp, rp_relPath rp )
        -- Mark the file as found
        -- TODO: if there is an error below, use a specific state
        modify $ SM.adjust ( _2 .~ True ) k
        seen <- gets $ liftM fst . SM.lookup k
        lift $ onErrorWarn $ do
          ( fid, samePosition ) <- case seen of
            Just fid @ ( FileId l _ ) | l == s && not withRescan -> do
              say Chatty $ "Skipping " ++ show ( rp_canPath rp )
              return ( fid, True )
            mfid -> do
              fid <- fileId $ rp_canPath rp
              lift $ ( \ a -> say Chatty a >> outputLineMsg a ) ( "Adding " ++ rp_canPath rp )
              return ( fid, Just fid == mfid )
          addFile rp samePosition fid
      -- set path priorities
      when withSetPri $ case mpri of
        Nothing -> return ()
        Just pri -> lift $ do
          withStorage ( rp_sid rp ) ( rp_sPath rp )
          setPath ( rp_relPath rp )
          addOp $ WithPriority pri
          addOp MarkPath


loadDB :: FSMonad m => ErrorT Error ( GenF4LMonad k s m ) BL.ByteString
loadDB = do
  f <- lift getDBFile
  processLengthAndBLFromFile' f $ flip const


dumpOps :: ( FSMonad m, BufferMonad m, Knowledge k ) => BL.ByteString -> TopF4LMonad k m ()
dumpOps = mapM_ ( outputMsg . show ) . readVOps


dumpFacts :: ( FSMonad m, BufferMonad m, Knowledge k ) => BL.ByteString -> TopF4LMonad k m ()
dumpFacts = loadOpsDo ( const $ return () ) ( outputMsg . show )


-- | This re-reads the DB file, so the cache could potentially be assigned a wrong hash
--   in case of concurrent modifications
saveCache :: ( DelFSMonad m, Knowledge k, SerializableKnowledge k ) => ErrorT Error ( TopF4LMonad k m ) ()
saveCache = do
  f <- lift $ getDBFile
  hs <- liftM ( map $ \ ( _, _, h ) -> h ) $ fromDB $ return . sessionsWithHash
  case reverse hs of
    [] -> return ()
    sha : past -> do
      k <- liftM kCode $ use st_knowledge
      let cfile s = combine ( takeDirectory f ) $ "cache-" ++ k ++ "." ++ hashToS s
      let cf = cfile sha
      p <- lift $ doesFileExist cf
      if p then say Timid ( "Cache file " ++ cf ++ " already exists" )
        else do
        opEnv <- use st_opEnv
        k <- use st_knowledge
        appendBLsToFile cf [ GZ.compress $ runPut $ B.put opEnv >> putK (opEnv ^. openv_fids) k ]
      -- Remove cache files for previous sessions
      lift $ mapM_ ( onErrorUseDefault () . removeFile . cfile ) past
  where
    noK :: Knowledge k1 => k1 -> k1
    noK = const noKnowledge


loadCached :: ( BufferMonad m, DelFSMonad m, ParFSMonad m, SerializableKnowledge k, B.Binary k ) => ErrorT Error ( TopF4LMonad k m ) ()
loadCached = do
  f <- lift $ getDBFile
  ess <- liftM ( map $ \ (_, a, b) -> (a, b) ) $ fromDB $ return . sessionsWithHash
  k <- liftM kCode $ use st_knowledge
  mc <- lift $ findCache ( takeDirectory f ) k [] $ reverse ess
  case mc of
    Just (cfile, bss) -> do
      let g = do
            openv <- B.get
            k <- getK ( openv ^. openv_fids )
            return ( openv, k )
      outputLineMsg "Loading cache. Please wait..."
      (opEnv, k) <- processLengthAndBLFromFile cfile $ const ( runGet g . GZ.decompress )
      opEnv `deepseq` st_opEnv .= opEnv
      st_knowledge .= k
      let lUncached = sum $ map ( fromIntegral . BL.length ) bss
      when ( lUncached > 1000000 ) $ lift $ fork $ onErrorWarn saveCache
      outputLineMsg "Loading ops"
      lift $ loadOps $ BL.concat bss
      cleanLine
    Nothing -> do
      fromDB loadOps
      lift $ fork $ onErrorWarn saveCache
      
  where
    cfile dir k h = combine dir $ "cache-" ++ k ++ "." ++ hashToS h
    findCache :: FSMonad m
                 => FilePath -> String -> [BL.ByteString] -> [( BL.ByteString, SBS.ShortByteString )] -> m ( Maybe ( FilePath, [BL.ByteString] ) )
    findCache _ _ _ [] = return Nothing
    findCache dir kCode pre ( ( bs, h ) : rest ) = do
      let f = cfile dir kCode h
      p <- doesFileExist f
      if p then return $ Just ( f, pre )
        else findCache dir kCode ( bs : pre ) rest


loadOpsDo :: ( FSMonad m, Knowledge k )
             => ( Op -> m () ) -> ( Fact -> m () ) -> BL.ByteString -> TopF4LMonad k m ()
loadOpsDo actOp actFact bs = do
  evalStateT ( lift ( return $ readOps bs ) >>= mapM_ apply >> lift cleanLine ) 0
  where
    apply op = do
      count <- get
      when ( count `mod` 1000 == 0 ) $ lift ( outputLineMsg $ hShow count ++ " ops loaded" )
      put $ count + 1
      facts <- lift $ execOp op
      lift $ lift $ actOp op
      lift $ flip mapM_ facts $ \ fact -> do
        st_knowledge %= applyFact fact
        lift $ actFact fact


sessions :: BL.ByteString -> [(SessionData, BL.ByteString)]
sessions dat = unfoldr getSession dat
  where
    getStartSession :: Get ( Maybe SessionData )
    getStartSession = do
      empty <- isEmpty
      if empty then return Nothing else do
        s <- B.get
        case s of
          StartSession a b c d e -> return $ Just $ SessionData a b c d e
          _ -> error $ "StartSession expected"

    getSession :: BL.ByteString -> Maybe ( ( SessionData, BL.ByteString ), BL.ByteString )
    getSession bs = case runGetOrFail getStartSession bs of
      Right ( ns, offset, Just dat ) ->
        let
          bsize = fromIntegral $ sdat_len dat
        in
         Just ( ( dat, BL.take ( offset + bsize ) bs ), BL.drop bsize ns )
      Right ( _, _, Nothing ) -> Nothing
      Left ( _, _, msg ) -> error msg


sessionsWithHash :: BL.ByteString -> [( SessionData, BL.ByteString, SBS.ShortByteString )]
sessionsWithHash s =
  flip evalState newSha1 $ flip mapM ( sessions s ) $ \ ( sd, bs ) -> do
    modify $ updateSha1 bs
    cs <- get
    return ( sd, bs, toSha1 cs )


fromDB :: FSMonad m => ( BL.ByteString -> GenF4LMonad k s m a ) -> ErrorT Error ( GenF4LMonad k s m ) a
fromDB a = loadDB >>= lift . a


loadOps :: ( FSMonad m, Knowledge k ) => BL.ByteString -> TopF4LMonad k m ()
loadOps = loadOpsDo ( const $ return () ) ( const $ return () )


readOps :: BL.ByteString -> [Op]
readOps = map toCurrentOp . readVOps


readVOps :: BL.ByteString -> [VOp]
readVOps s = concat $ map readSessionOps $ sessions s
  where
    readSessionOps (sd, bs) =
      case sdat_ver sd of
        V0 -> map VOp0 $ readMany bs
        V0_1 -> map VOp0_1 $ readMany bs


readMany :: B.Binary b => BL.ByteString -> [b]
readMany =  unfoldr getOne
  where
    mget :: B.Binary b => Get ( Maybe b )
    mget = do
      empty <- isEmpty
      if empty then return Nothing else liftM Just B.get
    getOne bs = case runGetOrFail mget bs of
      Right ( ns, _, Just op ) -> Just ( op, ns )
      Right ( _, _, Nothing ) -> Nothing
      Left ( _, _, msg ) -> error msg


readFullOps :: BL.ByteString -> [Op]
readFullOps s = r' ( readOps s ) initialOpEnv
  where
    r' [] _ = []
    r' ( op : rest ) opEnv =
      let
        newEnv = execState ( applyOp op ) opEnv
        oop = case newEnv `seq` op of
          WithKnownFid h ->
            case findFid h ( newEnv ^. openv_fids ) of
              [ fid ] -> WithFid fid
              _ -> error $ "Problem looking for known fid " ++ show h
          _ -> op 
      in oop : r' rest newEnv


sliceSessions :: ( FSMonad m, BufferMonad m ) => ( SessionData -> Bool ) -> [Op] -> TopF4LMonad () m ()
sliceSessions _ [] = return ()
sliceSessions cond ( StartSession ver sessid env ts l : ops ) =
  let
    dat = SessionData ver sessid env ts l
    isNotSession ( StartSession _ _ _ _ _ ) = False
    isNotSession _ = True
    ( these, rest ) = span isNotSession ops
  in
   case cond dat of
     False ->
       sliceSessions cond rest
     True -> do
       inSession' env ts sessid $ mapM_ procOp these
       sliceSessions cond rest
  where
    procOp :: BufferMonad m => Op -> F4LMonad () m ()
    procOp ( StartSession _ _ _ _ _ ) = error "StartSession within a session"
    procOp EndSession = return () -- Skip EndSession
    procOp ( WithFid fid ) = do
      use ( st_opEnv . to ( fidShorthand fid ) ) >>= \ sh -> case sh of
        Just s -> addOp $ WithKnownFid s
        Nothing -> addOp $ WithFid fid
    procOp op = addOp op
sliceSessions _ ( op : _ ) = error $ "Unexpected operator " ++ show op


importExportSessions :: ( FSMonad m, BufferMonad m ) 
                        => ( SessionData -> Bool ) -> FilePath -> FilePath -> F4LState () () -> m ()
importExportSessions cond ifile ofile st = do
  whenM ( doesFileExist ofile ) $ error $ "Not proceeding: file " ++ ofile ++ " already exists"
  idat <- onErrorDie $ processLengthAndBLFromFile' ifile $ flip const
  let ops = readFullOps idat
  evalStateT ( onErrorDie initDB >> sliceSessions cond ops ) ( conf_db .~ MetaDB ofile $ st )


stats :: ( FSMonad m, BufferMonad m, Knowledge k, FactCounterClass k, FileSetClass k ) 
         => GenF4LMonad k s m [(String, String)]
stats = do
  count <- use $ st_knowledge . fact_counter
  files <- use $ st_knowledge . fs_files
  return
    [ ( "facts", hShow $ count )
    , ( "files", hShow $ S.size files )
    , ( "total size", hShow $ totalSize files )
    ]
    where
      totalSize = sum . map fid_length . S.toList


storages :: ( FSMonad m, BufferMonad m, Knowledge k, SeenStoragesClass k ) 
            => GenF4LMonad k s m [(String, [String])]
storages = do
  m <- use $ st_knowledge . seen_storages
  return $ M.toList m


findStorage :: FSMonad m => StorageId -> [FilePath] -> m [FilePath]
findStorage sid paths = filterM isStorage paths
  where
    isStorage path = do
      esid <- runErrorT $ readFromFile $ combine path storageIdFile
      case esid of
        Right p | p == sid -> return True
        _ -> return False


storagesForEnv :: ( FSMonad m, KnownDataClass k )
                  => F4LMonad k m ( SM.Map StorageId [ FilePath ] )
storagesForEnv = do
  envid <- use st_envid
  say Timid $ "Env: " ++ envid
  -- Storages seen in this environment
  allStorages <- use $ st_knowledge . kd_storages
  let localStorages = storagesSeenIn envid allStorages
  liftM M.fromList $ mapM f $ M.toList localStorages
  where
    f (sid, paths) = liftM ( \ x -> (sid, x) ) $ findStorage sid paths


coverage :: ( FSMonad m, KnownDataClass k ) => GenF4LMonad k s m [(Int, Int, Float)]
coverage = do
  dat <- use $ st_knowledge . kd_files . to SM.elems
  let byCopies = foldl' ( \ a d -> SM.insertWith (+) (S.size $ fd_seen d) 1 a ) SM.empty dat
  let total = fromIntegral $ length dat
  return $ sort [(i, c, 100.0 * fromIntegral c / total) | (i, c) <- SM.toList byCopies]


labelStorage :: ( BufferMonad m, Knowledge k ) => String -> StorageId -> F4LMonad k m ()
labelStorage sid label = addOp $ AddStorageLabel sid label


labelEnv :: ( BufferMonad m, Knowledge k ) => String -> EnvironmentId -> F4LMonad k m ()
labelEnv envid label = addOp $ AddEnvironmentLabel envid label


-- | Get storage contents by initial part of storage id
listStorageContents :: ( FSMonad m, BufferMonad m ) => String -> TopF4LMonad KnownData m [([FilePath], FileId)]
listStorageContents psid = do
  onErrorDie $ fromDB loadOps
  fullData <- use st_knowledge
  let contents = view kd_contents fullData
  case M.toList $ M.filterWithKey (const . ( psid `isPrefixOf` )) contents of
    ( _sid, dat ) : [] ->
      return $ allFiles dat
    [] ->
      error "No storage with this ID found"
    xs ->
      error $ "Multiple storages found: " ++ show ( map fst xs )


autoBackup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k ) 
              => FilePath -> Int -> F4LMonad k m ()
autoBackup dest maxSize = do
  fullData <- use st_knowledge  
  -- Fids to backup by reverse weight
  let fids = map fst $ sortBy ( compare `byRev` benefitOverCost ) $ M.toList $ view kd_files fullData
  backup fids dest maxSize

  where
    -- TODO: find a more reasonable fun
    benefitOverCost :: ( FileId, FileData ) -> Float
    benefitOverCost ( fid, fd ) =
      ( fromIntegral $ pri $ fd_priority fd )
      * p ^^ ( S.size $ fd_seen fd ) * ( 1 - p )
      / fromIntegral ( max 1 $ fidDiskSpace fid )

    p = 0.2 -- Probability one copy of the file disappears

    pri :: Priority -> Int
    pri ( Priority i ) = 10 ^ i
    pri DefaultPriority = 1
    pri AsManyAsPossible = pri ( Priority 10 )
    pri _ = 0

    byRev :: ( a -> a -> c ) -> ( b -> a ) -> ( b -> b -> c )
    byRev f t = \ a b -> f ( t b ) ( t a )
  

backupPath :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
              => StorageId -> FilePath -> FilePath -> F4LMonad k m ()
backupPath sid path dest = do
  st <- use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid )
  let fids = map snd $ subFiles ( splitDirectories path ) st
  backup fids dest 1000000000000


backup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
          => [ FileId ] -> FilePath -> Int -> F4LMonad k m ()
backup fids dest maxSize = onErrorWarn $ do
  fullData <- lift $ use st_knowledge
  envid <- lift $ use st_envid
  say Timid $ "Env: " ++ envid
  storages <- lift $ storagesForEnv
  createDirectoryIfMissing True dest
  sid <- forceGetStorageId dest
  let seenfids = S.fromList $ ST.fids $ M.findWithDefault ST.empty sid $ view kd_contents fullData
  say Chatty $ "Seen fids: " ++ show seenfids
  let fidDataMap = view kd_files fullData
  let fidsData = map ( \ fid -> ( fid, M.findWithDefault noFileData fid fidDataMap ) ) fids
  let newfids = flip filter fidsData $ not . ( `S.member` seenfids ) . fst
  lift $ do
    withStorage sid dest
    runStateT ( backupFids storages maxSize newfids ) (0, 0)
    cleanLine

  where
    backupFids :: ( DelFSMonad m, BufferMonad m, Knowledge k ) => SM.Map StorageId [ FilePath ] -> Int -> [(FileId, FileData)] -> StateT (Int, Int) ( F4LMonad k m ) ()
    backupFids storages s fids = -- liftIO ( putStrLn $ "Backing up size: " ++ show s ) >>
      let nfids = dropWhile ( ( > s ) . fidDiskSpace . fst ) fids
      in
       case nfids of
         ( fid, dat ) : rest -> do
           let ( dp, destPath ) = destFidPath fid
           isThere <- doesFileExist destPath
           if isThere then lift ( say Chatty "Skipping" ) >> backupFids storages s rest else do
             lift $ onErrorWarn $ createDirectoryIfMissing True dp
             ( nfiles, donesize ) <- get
             let candidates = fileCandidates storages dat
             res <- lift $ runErrorT $ checkedCopyFileCandidates fid candidates destPath $ hShow nfiles ++ " " ++ hShow donesize ++ "\t"
             -- TODO: deal with disk full error
             case res of
               Right _ -> do
                 _1 += 1
                 _2 += fidDiskSpace fid
                 lift $ do
                   setFid fid
                   addOp BackupPath
                   addOp Seen
                 backupFids storages ( s - fidDiskSpace fid ) rest
               Left err -> do
                 -- say Quiet $ "Stopping on error: " ++ err
                 backupFids storages s rest
         _ ->
           return ()

    destFidPath fid =
      let ( p, fn ) = fidToBDest fid
      in ( combine dest p, combine dest $ combine p fn )


-- | Fail on errors with DB or destination
--   Report result: file id, destination path, flag
--   flag = Nothing -> file was already present
--   flag = Just True -> file restored
--   flag = Just False -> file not restored
restore :: ( DelFSMonad m, BufferMonad m, KnownDataClass k, Knowledge k ) 
           => Bool -> StorageId -> FilePath -> FilePath -> ErrorT Error ( F4LMonad k m ) [(FileId, FilePath, Maybe Bool)]
restore recorded sid srcpath dest = do
  createDirectoryIfMissing True dest
  -- Get destination storage data
  srp <- ( if recorded then forceGetStorage else fakeGetStorage ) dest
  lift $ do
    let destsid = rp_sid srp
    let sPath = rp_sPath srp
    let relDest = rp_relPath srp
    storages <- storagesForEnv
    let pfx = splitDirectories srcpath
    toCopy <- use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid ) . to ( subFiles pfx )
    say Timid $ show $ toCopy
    when recorded $ withStorage destsid dest
    kdFiles <- use $ st_knowledge . kd_files
    res <- mapM ( \ ( path, fid ) -> tryCopy storages kdFiles fid sPath $ joinPath $ relDest : path ) toCopy
    cleanLine
    return res
  where
    tryCopy :: ( DelFSMonad m, BufferMonad m, Knowledge k )
               => SM.Map StorageId [FilePath] -> SM.Map FileId FileData -> FileId -> FilePath -> FilePath 
               -> F4LMonad k m (FileId, FilePath, Maybe Bool)
    tryCopy storages files fid storagePath relDestPath = do
      let destPath = joinPath [ storagePath, relDestPath ]
      say Timid $ "Copying " ++ show fid ++ " to " ++ destPath
      outputLineMsg $ "Copying " ++ show fid ++ " to " ++ destPath
      isThere <- doesFileExist destPath
      case isThere of
        True -> do
          say Chatty "Skipping"
          return ( fid, relDestPath, Nothing )
        False -> do
          res <- runErrorT $ do
            createDirectoryIfMissing True $ takeDirectory destPath
            let mdat = M.lookup fid files
            case mdat of
              Just dat -> do
                let candidates = fileCandidates storages dat
                copied <- lift $ runErrorT $ checkedCopyFileCandidates fid candidates destPath ""
                case copied of
                  Right _ -> do
                    when recorded $ lift $ do
                      setFid fid
                      setPath relDestPath
                      addOp Seen
                    return ( fid, relDestPath, Just True )
                  Left _ -> do
                    say Chatty "Skipping: not found anywhere"
                    return ( fid, relDestPath, Just False )
              Nothing -> do
                say Chatty "Skipping: not found anywhere"
                return ( fid, relDestPath, Just False )
          return $ either ( const ( fid, relDestPath, Just False ) ) id res
