{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE GeneralizedNewtypeDeriving, FlexibleInstances #-}
module F4L.BufferMonad where

import F4L.FSMonad

import Control.Lens
-- import Control.Monad.IO.Class( MonadIO )
import Control.Monad.State.Strict
import qualified Control.Monad.State as SL
-- import Control.Monad.State( StateT, get, put )
-- import Control.Monad.Trans.Class
import Data.Array.IO
import Data.Array.Unboxed
import qualified Data.ByteString.Lazy as BL
-- import qualified Data.ByteString.Short as SBS
import Data.Word( Word8 )
import System.Console.Haskeline.MonadException
-- import System.IO( Handle )


class Monad m => BufferMonad m where
  emptyBuffer :: m ()
  addToBuffer :: BL.ByteString -> m ()
  freezedBuffer :: m (UArray Int Word8, Int)
  bufferLength :: m Int


instance BufferMonad m => BufferMonad ( StateT s m ) where
  emptyBuffer = lift emptyBuffer
  addToBuffer = lift . addToBuffer
  freezedBuffer = lift freezedBuffer
  bufferLength = lift bufferLength


type Buffer = Maybe (IOUArray Int Word8, Int)


growBuf :: Int -> IOWithBuffer ( IOUArray Int Word8 )
growBuf s = IOWithBuffer $ do
  mbuf <- get
  case mbuf of
    Nothing -> do
      buf <- lift ( newArray_ (0, s) ) 
      put $ Just (buf, 0)
      return buf
    Just (buf, c) -> do
      -- liftIO $ putStrLn $ "Buffer: growing buffer to " ++ show c
      nbuf <- lift $ newArray_ (0, s)
      lift $ sequence_ [ readArray buf i >>= writeArray nbuf i | i <- [0 .. c-1] ]
      put $ Just (nbuf, c)
      return nbuf


newtype IOWithBuffer a = IOWithBuffer { deIOWithBuffer :: StateT Buffer IO a }
                         deriving ( Monad, MonadIO, MonadState Buffer, FSMonad, DelFSMonad, ParFSMonad, MonadException )

runIOWithBuffer :: IOWithBuffer a -> IO a
runIOWithBuffer = flip evalStateT Nothing . deIOWithBuffer


-- | Extend MonadException to Lazy StateT
instance MonadException m => MonadException (SL.StateT s m) where
  controlIO f = SL.StateT $ \s -> controlIO $ \(RunIO run) -> let
    run' = RunIO (fmap (SL.StateT . const) . run . flip SL.runStateT s)
    in fmap (flip SL.runStateT s) $ f run'


traceBuffer :: IOWithBuffer ()
traceBuffer = do
  s <- get
  case s of
    Nothing -> outputMsg("Current buffer: None")
    Just (a, s) -> do
      l <- liftIO $ getElems a
      outputMsg("Current buffer(" ++ show s ++ "): " ++ show (take s l))


instance BufferMonad IOWithBuffer where
  emptyBuffer = put Nothing
  addToBuffer s = do
    -- liftIO $ putStrLn $ "Buffer: adding " ++ show s ++ " of length " ++ show (BL.length s)
    ms <- get
    case ms of
      Nothing -> do
        -- Initialize buffer and retry
        buf <- liftIO $ newArray_ (0, 1000)
        put $ Just (buf, 0)
        addToBuffer s
      Just ( buf, i ) -> do
        let sl = BL.unpack s
        (_, s) <- liftIO $ getBounds buf
        nbuf <- if i + length sl > s then
                  growBuf ( max (s * 3 `div` 2) (i + length sl) )
                else return buf
        liftIO $ sequence $ zipWith (writeArray nbuf) [i..] sl
        traverse . _2 += length sl
    -- traceBuffer
  freezedBuffer = do
    ms <- get
    case ms of
      Nothing -> return $ ( listArray (1,0) [], 0 )
      Just ( buf, l ) -> do
        fbuf <- liftIO $ freeze buf
        return $ ( fbuf, l )
  bufferLength = do
    get >>= return . maybe 0 snd
