{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.FSMonad where

import F4L( FileAttr, FilePathAttr( Property ) )
import F4L.Utils( uuidString )
import qualified F4L.Utils as U( fileSize )

import Control.Concurrent( forkIO )
--import qualified Control.Exception as EX
--import Control.Exception( bracket, tryJust, Exception )
import Control.Monad
import qualified Control.Monad.Catch as EX
import Control.Monad.Error( ErrorT(..), mapErrorT )
-- import qualified Control.Monad.Error as ERR
import Control.Exception( bracket, tryJust, Exception )
import Control.Monad.Instances() -- Provides monad for r ->
import Control.Monad.State.Strict( liftIO, StateT, evalStateT, get, put, runStateT )
import Control.Monad.Trans.Class
-- import Data.Array.Unboxed
import qualified Data.ByteString.Lazy as BL hiding ( hPutStr )
import qualified Data.ByteString as BS hiding ( hPutStr )
import Data.Maybe( catMaybes )
import Data.Time.Clock.POSIX( getPOSIXTime )
-- import Data.Word
import Graphics.Exif
import qualified System.Directory as Dir
import System.IO( openBinaryFile, hClose, hFileSize, hPutStr, stderr, withFile, IOMode( ReadMode, AppendMode, WriteMode ) )
import System.IO.Error( isDoesNotExistError )


type Error = String


class Monad m => FSMonad m where
  outputMsg :: String -> m ()
  outputLineMsg :: String -> m ()
  outputLineMsg _ = return ()
  cleanLine :: m ()
  cleanLine = return ()
  exifProperties :: FilePath -> m [FileAttr]
  generateUuid :: m String
  prompt :: String -> m String
  getNow :: m Int
  readFromFile :: FilePath -> ErrorT Error m String
  actOnLengthAndBLFromFile :: Integral i => FilePath -> ( i -> BL.ByteString -> m a ) -> ErrorT Error m a
  processLengthAndBLFromFile :: Integral i => FilePath -> ( i -> BL.ByteString -> a ) -> ErrorT Error m a
  processLengthAndBLFromFile f act = actOnLengthAndBLFromFile f ( liftM ( liftM return ) act )
  processLengthAndBLFromFile' :: Integral i => FilePath -> ( i -> BL.ByteString -> a ) -> ErrorT Error m a
  -- | Strict version. Read all file and close
  processLengthAndBLFromFile' = processLengthAndBLFromFile
  appendToFile :: FilePath -> String -> ErrorT Error m ()
  appendBLsToFile :: FilePath -> [BL.ByteString] -> ErrorT Error m ()
  -- | The default implementation is inefficient
  writeBLToFileAndDo :: FilePath -> a -> ( BS.ByteString -> a -> m a ) -> BL.ByteString -> ErrorT Error m a
  writeBLToFileAndDo f a act bl = do
    appendBLsToFile f [bl]
    lift $ foldM ( flip act ) a $ BL.toChunks bl
  canonicalizePath :: FilePath -> m FilePath
--  copyFile :: FilePath -> FilePath -> ErrorT Error m ()
  doesDirectoryExist :: FilePath -> m Bool
  doesFileExist :: FilePath -> m Bool
  fileSize :: Integral i => FilePath -> ErrorT Error m i
  getCurrentDirectory :: m FilePath
  getDirectoryContents :: FilePath -> ErrorT Error m [FilePath]
  createDirectoryIfMissing :: Bool -> FilePath -> ErrorT Error m ()
  myCatch :: Exception e => m a -> ( e -> m a ) -> m a


class FSMonad m => DelFSMonad m where
  removeFile :: FilePath -> ErrorT Error m ()


class FSMonad m => ParFSMonad m where
  fork :: m () -> m ()


instance FSMonad m => FSMonad ( StateT s m ) where
  outputMsg = lift . outputMsg
  outputLineMsg = lift . outputLineMsg
  cleanLine = lift cleanLine
  exifProperties = lift . exifProperties
  generateUuid = lift generateUuid
  prompt = lift . prompt
  getNow = lift $ getNow
  readFromFile = mapErrorT lift  . readFromFile
  actOnLengthAndBLFromFile f act = do
    s <- get
    (a, ns) <- mapErrorT lift $ actOnLengthAndBLFromFile f ( \ i dat -> flip runStateT s $ act i dat )
    put ns
    return a
  processLengthAndBLFromFile = liftM ( liftM $ mapErrorT lift ) processLengthAndBLFromFile
  processLengthAndBLFromFile' = liftM ( liftM $ mapErrorT lift ) processLengthAndBLFromFile'
  appendToFile fp s = mapErrorT lift $ appendToFile fp s
  appendBLsToFile = liftM ( liftM $ mapErrorT lift ) appendBLsToFile
  writeBLToFileAndDo p a act bl = do
    s <- get
    (a, ns) <- mapErrorT lift $ writeBLToFileAndDo p ( a, s ) ( unlift act ) bl
    put ns
    return a
    where
      unlift :: ( BS.ByteString -> a -> StateT s m a ) -> BS.ByteString -> ( a, s ) -> m ( a, s )
      unlift f bs (a, s) = flip runStateT s $ f bs a
  canonicalizePath = lift . canonicalizePath
--  copyFile = liftM ( liftM $ mapErrorT lift ) copyFile
  doesDirectoryExist = lift . doesDirectoryExist
  doesFileExist = lift . doesFileExist
  fileSize = mapErrorT lift . fileSize
  getCurrentDirectory = lift getCurrentDirectory
  getDirectoryContents = mapErrorT lift . getDirectoryContents
  createDirectoryIfMissing = liftM ( liftM $ mapErrorT lift ) createDirectoryIfMissing
  myCatch act proc = do
    s <- get
    res <- lift $ myCatch (runStateT act s >>= return . Right) (return . Left)
    case res of
      Right (a, ns) -> do
        put ns
        return a
      Left e ->
        proc e


instance DelFSMonad m => DelFSMonad ( StateT s m ) where
  removeFile = mapErrorT lift . removeFile


instance ParFSMonad m => ParFSMonad ( StateT s m ) where
  fork a = do
    s <- get
    lift $ fork $ evalStateT a s


ioCatch :: IO a -> ErrorT Error IO a
ioCatch a = ErrorT $ EX.handleIOError ( return . Left . show ) $ liftM Right a


instance FSMonad IO where
  outputMsg = putStrLn

--  outputLineMsg msg = hPutStr stderr "\ESC[0G\ESC[2K" >> putStr msg
  -- TODO: use console size
  outputLineMsg msg = hPutStr stderr $ "\r\ESC[K" ++ take 79 msg

  cleanLine = outputLineMsg "\r"

  exifProperties path = do
    mexif <- liftIO $ EX.catch ( liftM Just $ fromFile path ) skipEx
    liftM catMaybes $ sequence $ map ( tagToAttr mexif ) ["DateTime"]
    where
      skipEx :: EX.SomeException -> IO ( Maybe a )
      skipEx _ = return Nothing
      tagToAttr Nothing _ = return Nothing
      tagToAttr ( Just exif ) tag = do
        mval <- EX.catchIOError ( getTag exif tag ) ( const $ return Nothing )
        return $ liftM ( \ v -> Property ( "exif:" ++ tag ) v ) mval

  prompt msg = do
    putStrLn msg
    getLine

  generateUuid = uuidString
  
  readFromFile = ioCatch . readFile

  getNow = liftM round getPOSIXTime

  actOnLengthAndBLFromFile f action =
    ioCatch $ bracket (openBinaryFile f ReadMode) hClose $ \h -> do
      size <- hFileSize h
      dat <- BL.hGetContents h
      res <- action ( fromIntegral size ) dat
      res `seq` return res

  processLengthAndBLFromFile f action = actOnLengthAndBLFromFile f ( \ a b -> return $ action a b )

  processLengthAndBLFromFile' f action =
    ioCatch $ bracket (openBinaryFile f ReadMode) hClose $ \h -> do
      size <- hFileSize h
      liftM ( action $ fromIntegral size ) $ BL.hGet h $ fromIntegral size

  appendToFile = liftM ( liftM ioCatch ) appendFile
  appendBLsToFile f bls = ioCatch $ withFile f AppendMode $ \ h -> do
    mapM_ ( BL.hPut h ) bls
  writeBLToFileAndDo f a act bl = ioCatch $ withFile f WriteMode $ \ h -> do
    foldM ( act' h ) a $ BL.toChunks bl
    where act' h a s = do
            BS.hPut h s
            act s a
  canonicalizePath p = do
    cp <- tryJust ( guard . isDoesNotExistError ) $ Dir.canonicalizePath p
    return $ either (const p) id cp
--  copyFile = liftM ( liftM ioCatch ) Dir.copyFile
  doesDirectoryExist = Dir.doesDirectoryExist
  doesFileExist = Dir.doesFileExist
  fileSize = ioCatch . U.fileSize
  getCurrentDirectory = Dir.getCurrentDirectory
  getDirectoryContents = ioCatch . Dir.getDirectoryContents
  createDirectoryIfMissing = liftM ( liftM ioCatch ) Dir.createDirectoryIfMissing
  myCatch = EX.catch


instance DelFSMonad IO where
  removeFile = ioCatch . Dir.removeFile


instance ParFSMonad IO where
  fork a = forkIO a >> return ()
