{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE MultiParamTypeClasses, FlexibleContexts, RankNTypes #-}

module F4L.Op (
    Fact(..)
  , FidStore
  , Op(..)
  , OpEnv(..)
  , VOp(..)
  , applyOp
  , fidShorthand
  , findFid
  , findShorthand
  , initialOpEnv
  , mfact_file
  , mfact_storage
  , mfact_storagepath
  , openv_envid
  , openv_envid'
  , openv_fid
  , openv_fids
  , openv_path
  , openv_storage
  , openv_ts
  , openv_ts'
  , toCurrentOp
  , withFid
  ) where

import F4L
import qualified F4L.V0.Op as V0_Op
import F4L.Utils( fromJust', getList, getStrict )

import Control.DeepSeq
import Control.Lens
import Control.Monad.State.Strict
import qualified Data.Binary as B
import qualified Data.ByteString as BS
import qualified Data.ByteString.Short as SBS
-- import qualified Data.ByteString.Lazy as BL
import Data.Binary( Binary )
import Data.List( find )
import Data.Map.Strict( Map )
import qualified Data.Map.Strict as M
-- import Data.Map( Map )
-- import qualified Data.Map as M
import Data.Maybe( maybeToList, mapMaybe )
import Data.Set( Set )
import qualified Data.Set as S
-- import Data.Trie( Trie )
-- import qualified Data.Trie as T
import Data.Word( Word8 )
import System.FilePath( combine )


data Op = StartSession SessionVersion SessionId EnvironmentId TimeStamp Int
        | EndSession

        | AddEnvironmentLabel EnvironmentId String
        | AddStorageLabel StorageId String

        | WithStorage StorageId FilePath

        | WithTag String -- ^ add tag for next file or path
        | WithPriority Priority -- ^ set priority for following path
        | WithProperty String String -- ^ add property for next file or path

        | RetailPath Int FilePath -- ^ remove n chars and append new part
        | PatchPath Int Int FilePath -- ^ cut from pos1 to pos2 and replace with new part
        | BackupPath -- ^ path within storage is standard backup path
        
        | WithFid FileId
        | WithKnownFid SBS.ShortByteString
        | MarkPath -- ^ add attributes to storage + path

        | Seen -- ^ mark path as seen
        | Mark -- ^ add attributes
        | SeenMark -- ^ Seen + Mark
        | Missing -- ^ no file present in path

        | Nop -- ^ no operation
        deriving (Show, Eq)


-- type FidStore = Trie FileId -- ^ hash to fid

type FidStore = Map SBS.ShortByteString FileId


{-
instance Binary a => Binary ( Trie a ) where
  put = B.put . T.toList
  get = liftM T.fromList B.get
-}


data OpEnv = OpEnv { _openv_envid :: !( Maybe EnvironmentId )
                   , _openv_ts :: !( Maybe TimeStamp )
                   , _openv_storage :: !( Maybe ( StorageId, FilePath ) )
                   , _openv_priority :: !( Maybe Priority )
                   , _openv_fid :: !( Maybe FileId )
                   , _openv_path :: !FilePath
                   , _openv_tags :: !( Set String )
                   , _openv_props :: !( Map String String )
                   , _openv_fids :: !FidStore -- ^ hash to fid
                   }
             deriving Show

openv_envid :: Lens' OpEnv ( Maybe EnvironmentId )
openv_envid = lens _openv_envid ( \ s e -> s { _openv_envid = e } )

just_lens :: String -> Lens' ( Maybe a ) a
just_lens label = lens ( fromJust' $ "Expected " ++ label ) ( const Just )

openv_envid' :: Lens' OpEnv EnvironmentId
openv_envid' = openv_envid . just_lens "EnvironmentId"

openv_ts :: Lens' OpEnv ( Maybe TimeStamp )
openv_ts = lens _openv_ts ( \ s a -> s { _openv_ts = a } )

openv_ts' :: Lens' OpEnv TimeStamp
openv_ts' = openv_ts . just_lens "TimeStamp"

openv_storage :: Lens' OpEnv ( Maybe ( StorageId, FilePath ) )
openv_storage = lens _openv_storage ( \ s a -> s { _openv_storage = a } )

openv_priority :: Lens' OpEnv ( Maybe Priority )
openv_priority = lens _openv_priority  ( \ s a -> s { _openv_priority  = a } )

openv_fid :: Lens' OpEnv ( Maybe FileId )
openv_fid  = lens _openv_fid  ( \ s a -> s { _openv_fid  = a } )

openv_path :: Lens' OpEnv FilePath
openv_path = lens _openv_path  ( \ s a -> a `deepseq` s { _openv_path  = a } )

openv_tags = lens _openv_tags  ( \ s a -> s { _openv_tags  = a } )
openv_props = lens _openv_props  ( \ s a -> s { _openv_props  = a } )

openv_fids :: Lens' OpEnv FidStore
openv_fids = lens _openv_fids  ( \ s a -> s { _openv_fids  = a } )

openv_attrs :: Bool -> Getter OpEnv ( Maybe [FilePathAttr a] )
openv_attrs withPri = to attrs
  where
    attrs :: OpEnv -> Maybe [FilePathAttr a]
    attrs s = case attrs' s of
      [] -> Nothing
      x -> Just x
    attrs' s =
      if withPri then ( maybeToList $ liftM Require $ _openv_priority s ) else []
      ++ map Tag ( S.toList $ _openv_tags s )
      ++ map ( uncurry Property ) ( M.toList $ _openv_props s )


instance Binary OpEnv where
  put ( OpEnv a b c d e f g h i ) = B.put a >> B.put b >> B.put c >> B.put d >> B.put e >> B.put f >> B.put g >> B.put h >> B.put i
  get = do
    a <- B.get
    b <- B.get
    c <- B.get
    d <- B.get
    e <- B.get
    f <- B.get
    g <- B.get
    h <- B.get
    i <- liftM M.fromList $ getList getStrict
    return $ OpEnv a b c d e f g h i


instance NFData OpEnv where
  rnf ( OpEnv a b c d e f g h i ) = rnf a `seq` rnf b `seq` rnf c `seq` rnf d `seq` rnf e `seq` rnf f `seq` rnf g `seq` rnf h `seq` rnf i

instance NFData Op where
  rnf ( StartSession _ a b _ _ ) = rnf a `seq` rnf b
  rnf ( AddEnvironmentLabel a b ) = rnf a `seq` rnf b
  rnf ( AddStorageLabel a b ) = rnf a `seq` rnf b
  rnf ( WithStorage a b ) = rnf a `seq` rnf b
  rnf ( WithPriority a ) = rnf a
  rnf ( RetailPath a b ) = a `seq` rnf b
  rnf ( PatchPath a b c ) = a `seq` b `seq` rnf c
  rnf ( WithFid a ) = rnf a
  rnf ( WithKnownFid a ) = rnf a
  rnf ( WithTag a ) = rnf a
  rnf ( WithProperty a b ) = rnf a `seq` rnf b
  rnf x = x `seq` ()

{-
findShorthand :: FileId -> FidStore -> Maybe SBS.ShortByteString
findShorthand fid @ ( FileId _ h ) fids =
  liftM SBS.toShort $ find isUnique $ drop 1 $ BS.inits $ SBS.fromShort h
    where
      isUnique s =
        let sm = T.submap s fids
        in T.size sm == 1 && T.elems sm == [fid]

findFid :: SBS.ShortByteString -> FidStore -> [ FileId ]
findFid short fids = T.elems ( T.submap ( SBS.fromShort short ) fids )

withFid :: FileId -> FidStore -> FidStore
withFid fid fids =
  T.insert ( SBS.fromShort $ fid_hash fid ) fid fids
-}


findShorthand :: FileId -> FidStore -> Maybe SBS.ShortByteString
findShorthand fid @ ( FileId _ h ) fids = find cond $ map SBS.toShort $ drop 1 $ BS.inits $ SBS.fromShort h
  where
    cond short = M.lookup short fids == Just fid

findFid :: SBS.ShortByteString -> FidStore -> [ FileId ] -- TODO: return Maybe FileId
findFid short fids = take 1 $ mapMaybe cond $ reverse $ drop 1 $ BS.inits short'
  where
    short' = SBS.fromShort short
    cond sh = do
      x @ ( FileId _ h ) <- M.lookup ( SBS.toShort sh ) fids
      if BS.take ( SBS.length short ) ( SBS.fromShort h ) == short' then return x else Nothing

withFid :: FileId -> FidStore -> FidStore
withFid fid @ ( FileId _ h ) fids = withFid' ( drop 1 $ BS.inits ( SBS.fromShort h ) )
  where
    withFid' [] = error $ "Exausted hash length while looking for slot matching " ++ show h
    withFid' ( x : xs ) =
      let sh = SBS.toShort x in 
      case M.lookup sh fids of
        Nothing -> M.insert sh fid fids
        Just fid' | fid == fid' -> fids
        Just ( FileId _ h' ) | h == h' -> 
          error $ "Found a SHA1 collision for " ++ show fid ++ "?"
        _ -> withFid' xs


{-
type FidStore = Trie Int -- ^ hash to length

findShorthand :: FileId -> FidStore -> Maybe BS.ByteString
findShorthand fid @ ( FileId l h ) fids = 
  find isUnique $ BS.inits h
    where
      isUnique s =
        let sm = T.submap s fids
        in T.size sm == 1 && T.elems sm == [l]

findFid :: BS.ByteString -> FidStore -> [ FileId ]
findFid short fids = map ( uncurry $ flip FileId ) $ T.toList ( T.submap short fids )

withFid :: FileId -> FidStore -> FidStore
withFid fid fids =
  T.insert ( fid_hash fid ) ( fid_length fid ) fids
-}

initialOpEnv :: OpEnv
initialOpEnv = OpEnv Nothing Nothing Nothing Nothing Nothing "" S.empty M.empty M.empty

fidShorthand :: FileId -> OpEnv -> Maybe SBS.ShortByteString
fidShorthand fid env = findShorthand fid $ view openv_fids env

-- opEnvModify f = modify f >> return []

-- applyOp' op = applyOp $!! op

applyOp :: MonadState OpEnv m => Op -> m [Fact]

applyOp Nop = return []

applyOp ( StartSession _ _ envid ts _ ) = do
  fids <- use openv_fids
  put initialOpEnv
  openv_fids .= fids
  openv_envid .= ( Just $!! envid )
  openv_ts .= ( Just $!! ts )
  return []

applyOp EndSession = do
  fids <- use openv_fids
  put initialOpEnv
  openv_fids .= fids
  return []

applyOp ( AddEnvironmentLabel env tag ) = return [ LabelEnvironment env tag ]
applyOp ( AddStorageLabel env tag ) = return [ LabelStorage env tag ]

applyOp ( WithStorage sid path ) = do 
  openv_storage .= ( Just $!! ( sid, path ) )
  openv_path .= ""
  envid <- use openv_envid'
  ts <- use openv_ts'
  return [ StorageSeen ts envid sid path ]

applyOp ( WithPriority req ) = openv_priority .= ( Just $!! req ) >> return []

applyOp BackupPath = do
  mfid <- use openv_fid
  case mfid of
    Just fid ->
      let ( dir, fn ) = fidToBDest fid
      in openv_path .= combine dir fn
    Nothing ->
      error "Trying to set BackupPath with unset file Id"
  return []
applyOp ( RetailPath n append ) = openv_path %= retail >> return []
  where retail path = ( reverse $ drop ( fromIntegral n ) $ reverse path ) ++ append
applyOp ( PatchPath d k replace ) = openv_path %= patch >> return []
  where patch path = force $ ( reverse $ drop ( fromIntegral d ) $ reverse path ) ++ replace ++ reverse ( take k $ reverse path )
applyOp ( WithFid fid ) = do
  openv_fid .= ( Just $!! fid )
  openv_fids %= withFid fid
  return []
applyOp ( WithKnownFid h ) = do -- TODO: handle errors setting fid to none
  found <- use $ openv_fids . to ( findFid h )
  case found of
    [ fid ] -> openv_fid .= ( Just $!! fid )
    [] -> error $ "No known fid with prefix " ++ show h
    xs -> error $ "Multiple fids with prefix " ++ show h ++ ": " ++ show xs
  return []
applyOp ( WithTag s ) = openv_tags %= S.insert s >> return []
applyOp ( WithProperty k v ) = openv_props %= M.insert k v >> return []

applyOp MarkPath = do 
  s <- get
  openv_priority .= Nothing
  openv_tags .= S.empty
  openv_props .= M.empty
  return $ maybeToList $ liftM3 ( PathAttr $ s ^. openv_ts' ) ( liftM fst $ s ^. openv_storage ) ( Just $ s ^. openv_path ) ( s ^. openv_attrs True )

applyOp Seen = do
  s <- get
  return $ maybeToList $ liftM3 ( FileSeen $ s ^. openv_ts' ) ( s ^. openv_fid ) ( liftM fst $ s ^. openv_storage ) ( Just $ s ^. openv_path )

applyOp Mark = do
  s <- get
  openv_priority .= Nothing
  openv_tags .= S.empty
  openv_props .= M.empty
  return $ maybeToList $ liftM2 ( FileAttr $ s ^. openv_ts' ) ( s ^. openv_fid ) ( s ^. openv_attrs False )

applyOp SeenMark = liftM2 (++) ( applyOp Seen ) ( applyOp Mark )

applyOp Missing = do
  s <- get
  return $ maybeToList $ liftM2 ( FileMiss $ s ^. openv_ts' ) ( liftM fst $ s ^. openv_storage ) ( Just $ s ^. openv_path )


data Fact = FileSeen { fact_ts :: !TimeStamp
                     , fact_file :: !FileId
                     , fact_storage :: !StorageId
                     , fact_path :: !FilePath -- ^ path below storage where file was seen
                     } 
          | FileMiss { fact_ts :: !TimeStamp
                     , fact_storage :: !StorageId
                     , fact_path :: !FilePath -- ^ path below storage where no registered file was present
                     }
          | FileAttr { fact_ts :: !TimeStamp
                     , fact_file :: !FileId
                     , fact_file_attrs :: ![FileAttr]
                     }
          | PathAttr { fact_ts :: !TimeStamp
                     , fact_storage :: !StorageId
                     , fact_path :: !FilePath -- ^ path below storage where directory is
                     , fact_path_attrs :: ![PathAttr]
                     }
          | StorageSeen { fact_ts :: !TimeStamp
                        , fact_env :: !EnvironmentId
                        , fact_storage :: !StorageId
                        , fact_path :: !FilePath
                        }
          | LabelEnvironment { fact_env :: !EnvironmentId
                             , fact_tag :: !String
                             }
          | LabelStorage { fact_storage :: !StorageId
                         , fact_tag :: !String
                         }
          deriving Show

mfact_file :: Fact -> Maybe FileId
mfact_file ( FileMiss _ _ _ ) = Nothing
mfact_file ( PathAttr _ _ _ _ ) = Nothing
mfact_file ( StorageSeen _ _ _ _ ) = Nothing
mfact_file ( LabelEnvironment _ _ ) = Nothing
mfact_file ( LabelStorage _ _ ) = Nothing
mfact_file x = Just $ fact_file x

mfact_storage :: Fact -> Maybe StorageId
mfact_storage ( FileAttr _ _ _ ) = Nothing
mfact_storage ( StorageSeen _ _ _ _ ) = Nothing
mfact_storage ( LabelEnvironment _ _ ) = Nothing
mfact_storage ( LabelStorage _ _ ) = Nothing
mfact_storage x = Just $ fact_storage x

-- TODO: change name
mfact_storagepath :: Fact -> Maybe (StorageId, FilePath)
mfact_storagepath ( FileAttr _ _ _ ) = Nothing
mfact_storagepath ( StorageSeen _ _ _ _ ) = Nothing
mfact_storagepath ( LabelEnvironment _ _ ) = Nothing
mfact_storagepath ( LabelStorage _ _ ) = Nothing
mfact_storagepath x = Just ( fact_storage x, fact_path x )

instance NFData Fact where
  rnf ( FileSeen _ a b c ) = rnf a `seq` rnf b `seq` rnf c
  rnf ( FileMiss _ a b ) = rnf a `seq` rnf b
  rnf ( FileAttr _ a b ) = rnf a `seq` rnf b
  rnf ( PathAttr _ a b c ) = rnf a `seq` rnf b `seq` rnf c
  rnf ( StorageSeen _ a b c ) = rnf a `seq` rnf b `seq` rnf c
  rnf ( LabelEnvironment a b ) = rnf a `seq` rnf b
  rnf ( LabelStorage a b ) = rnf a `seq` rnf b


data VOp = VOp0 !V0_Op.Op
         | VOp0_1 !Op
         deriving( Eq, Show )


{---------------------------------------- Conversions ----------------------------------------}

toCurrentOp :: VOp -> Op
toCurrentOp ( VOp0_1 op ) = op
toCurrentOp ( VOp0 op ) = fromV0 op
  where
    fromV0 :: V0_Op.Op -> Op
    fromV0 ( V0_Op.StartSession _ a b c _ ) = StartSession currentVersion a b c undefined
    fromV0 V0_Op.EndSession = EndSession
    fromV0 ( V0_Op.AddEnvironmentLabel a b ) = AddEnvironmentLabel a b
    fromV0 ( V0_Op.AddStorageLabel a b ) = AddStorageLabel a b
    fromV0 ( V0_Op.WithStorage a b ) = WithStorage a b
    fromV0 ( V0_Op.WithRequire ( V0_Op.MergePriority a ) ) = WithPriority a
    fromV0 ( V0_Op.WithRequire ( V0_Op.OverridePriority a ) ) = WithPriority a
    fromV0 ( V0_Op.MTag _ ) = Nop
    fromV0 ( V0_Op.RetailPath a b ) = RetailPath a b
    fromV0 ( V0_Op.PatchPath a b c ) = PatchPath a b c
    fromV0 V0_Op.BackupPath = BackupPath
    fromV0 ( V0_Op.WithFid a ) = WithFid a
    fromV0 ( V0_Op.WithKnownFid a ) = WithKnownFid a
    fromV0 ( V0_Op.WithTag a ) = WithTag a
    fromV0 ( V0_Op.WithProperty a b ) = WithProperty a b
    fromV0 V0_Op.Seen = Seen
    fromV0 V0_Op.Mark = Mark
    fromV0 V0_Op.SeenMark = SeenMark
    fromV0 V0_Op.Missing = Missing


{---------------------------------------- Binary ----------------------------------------}

swap (a, b) = (b, a)

freqTags = M.fromList $
           [ ( 'p', "picture" )
           , ( 'm', "email" )
           ]
freqTags' = M.fromList $ map swap $ M.toList freqTags

freqProps = M.fromList $
           [ ( 'D', "exif:DateTime" )
           ]
freqProps' = M.fromList $ map swap $ M.toList freqProps


instance Binary Op where
  put ( StartSession sver sessid env ts l ) = B.put 'S' >> B.put sver >> B.put sessid >> B.put env >> B.put ts >> B.put l
  put EndSession = B.put 'E'
  put ( AddEnvironmentLabel env tag ) = B.put 'L' >> B.put env >> B.put tag
  put ( AddStorageLabel env tag ) = B.put 'l' >> B.put env >> B.put tag
  put ( WithStorage sid path ) = B.put 's' >> B.put sid >> B.put path
  put ( WithPriority pri ) = B.put 'r' >> B.put pri
  put BackupPath = B.put 'b'
  put ( RetailPath n path ) = B.put 'p' >> B.put ( fromIntegral n :: Word8 ) >> B.put path
  put ( PatchPath d k path ) = B.put 'P' >> B.put ( fromIntegral d :: Word8 ) >> B.put ( fromIntegral k :: Word8 ) >> B.put path
  put ( WithFid fid ) = B.put 'i' >> B.put fid
  put ( WithKnownFid h ) = B.put 'I' >> B.put ( SBS.unpack h )
  put ( WithTag s ) = case M.lookup s freqTags' of
    Nothing ->
      B.put 't' >> B.put s
    Just x ->
      B.put '_' >> B.put x
  put ( WithProperty k v ) = case M.lookup k freqProps' of
    Nothing ->
      B.put 'a' >> B.put k >> B.put v
    Just x ->
      B.put '^' >> B.put x >> B.put v
  put Seen = B.put '1'
  put Mark = B.put 'm'
  put SeenMark = B.put 'M'
  put MarkPath = B.put 'R'
  put Missing = B.put '0'
  put Nop = return ()

  get = do
    t <- B.get
    case t of
      'S' -> liftM5 StartSession B.get B.get B.get B.get B.get
      'E' -> return EndSession
      'L' -> liftM2 AddEnvironmentLabel B.get B.get
      'l' -> liftM2 AddStorageLabel B.get B.get
      's' -> liftM2 WithStorage B.get B.get
      'r' -> liftM WithPriority B.get
      -- 'T' -> liftM MTag B.get
      'b' -> return BackupPath
      'p' -> liftM2 RetailPath (liftM (fromIntegral :: Word8 -> Int) B.get) B.get
      'P' -> liftM3 PatchPath (liftM (fromIntegral :: Word8 -> Int) B.get) (liftM (fromIntegral :: Word8 -> Int) B.get) B.get
      'i' -> liftM WithFid B.get
      'I' -> liftM ( WithKnownFid . SBS.pack ) B.get
      't' -> liftM WithTag B.get
      '_' -> liftM ( WithTag . flip ( M.findWithDefault "-" ) freqTags ) B.get
      'a' -> liftM2 WithProperty B.get B.get
      '^' -> liftM2 ( WithProperty . flip ( M.findWithDefault "-" ) freqProps ) B.get B.get
      '1' -> return Seen
      'm' -> return Mark
      'M' -> return SeenMark
      'R' -> return MarkPath
      '0' -> return Missing
      x -> error $ "Wrong tag: (code " ++ show ( fromEnum x ) ++ ")"
