{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.SeenTree (
    SeenNode(..)
  , SeenTree
  , allFiles
  , empty
  , fids
  , findFileNode
  , findNode
  , fromLazy
  , getSeenTree
  , insertFile
  , removeFile
  , subFiles
  , toLazy
  ) where

import F4L
import F4L.Utils

import Control.DeepSeq
import Control.Monad
import Data.Binary
import qualified Data.Map as M
import Data.Map.Strict hiding ( map, null, empty )
import qualified Data.Map.Strict as SM
-- import qualified Data.Text as T
-- import Data.Text( Text )
import System.FilePath( splitDirectories )


data SeenNode a = File !a
                | Dir !( SM.Map ShortPath ( SeenNode a ) )
                deriving (Eq, Ord, Show)

data LazyNode a = LazyFile a | LazyDir ( M.Map ShortPath ( LazyNode a ) )

toLazy :: SeenNode a -> LazyNode a
toLazy ( File a ) = LazyFile a
toLazy ( Dir d ) = LazyDir $ M.map toLazy $ M.fromList $ SM.toList d


fromLazy :: LazyNode a -> SeenNode a
fromLazy ( LazyFile a ) = File a
fromLazy ( LazyDir d ) = Dir $ SM.fromList $ M.toList $ M.map fromLazy d


instance Functor SeenNode where
  fmap f ( File a ) = File ( f a )
  fmap f ( Dir d ) = Dir $ SM.map ( fmap f ) d

instance Functor LazyNode where
  fmap f ( LazyFile a ) = LazyFile ( f a )
  fmap f ( LazyDir d ) = LazyDir $ M.map ( fmap f ) d

type SeenTree = SeenNode FileId

{-
instance Binary Text where
  put = put . T.unpack
  get = liftM T.pack get
-}

instance ( Binary a, NFData a ) => Binary ( SeenNode a ) where
  put ( File a ) = a `seq` put 'F' >> put a
  put ( Dir a ) = a `seq` put 'D' >> put a
  get = do
    c <- get
    case c of
      'F' -> liftM ( File $!! ) get
      'D' -> do
        m <- get
        rnf m `seq` return ( Dir m )

instance NFData a => NFData ( SeenNode a ) where
  rnf ( File a ) = rnf a
  rnf ( Dir sm ) = rnf sm

instance ( Binary a ) => Binary ( LazyNode a ) where
  put ( LazyFile a ) = put 'F' >> put a
  put ( LazyDir a ) = put 'D' >> put a
  get = do
    c <- get
    case c of
      'F' -> liftM LazyFile get
      'D' -> liftM LazyDir get

instance NFData a => NFData ( LazyNode a ) where
  rnf ( LazyFile a ) = rnf a
  rnf ( LazyDir sm ) = rnf sm

getSeenTree :: Binary a => ( a -> FileId ) -> Get SeenTree
getSeenTree f = do
  c <- get
  case c of
    'F' -> liftM ( \ a -> File $!! f a ) get
    'D' -> liftM ( Dir $!! ) $ getMap getStrict $ getSeenTree f


empty :: SeenTree
empty = Dir SM.empty


insertFile :: FilePath -> FileId -> SeenTree -> SeenTree
insertFile fp = insertFile' ( splitDirectories fp )

insertFile' [] dat _ = File dat
insertFile' x @ ( _ : _ )  dat ( File _ ) = insertFile' x dat empty
insertFile' ( x : xs ) dat ( Dir d ) = Dir $! insertWith ( const ins ) ( pathToShort x ) ( ins empty ) d
  where ins = insertFile' xs dat


removeFile :: FilePath -> SeenTree -> SeenTree
removeFile fp = removeFile' ( splitDirectories fp )

removeFile' [] _ = empty
removeFile' ( x : xs ) pre @ ( File _ ) = pre
removeFile' ( x : xs ) ( Dir d ) = Dir $ adjust ( removeFile' xs ) ( pathToShort x) d


fids :: SeenTree -> [FileId]
fids ( File fid ) = return fid
fids ( Dir d ) = concatMap fids $ SM.elems d


subDirs :: SeenTree -> [FilePath]
subDirs ( Dir st ) = do
  (p, v) <- SM.toList st
  case v of
    Dir _ -> return $ shortToPath p
    _ -> []
subDirs _ = []

allFiles :: SeenNode FileId -> [([FilePath], FileId)]
allFiles ( File f ) = return ([], f)
allFiles ( Dir st ) = do 
  ( p, node ) <- SM.toList st
  ( path, fid ) <- allFiles node
  return ( shortToPath p : path, fid )


findNode :: [FilePath] -> SeenTree -> Maybe SeenTree
findNode [] st = Just st
findNode _ ( File _ ) = Nothing 
findNode ( x : xs ) ( Dir st ) = SM.lookup ( pathToShort x ) st >>= findNode xs


findFileNode :: FilePath -> SeenTree -> Maybe FileId
findFileNode path st =
  case findNode ( splitDirectories path ) st of
    Just ( File x ) -> Just x
    _ -> Nothing

-- | list of (relative subpaths, fileid) - pre is omitted
subFiles :: [FilePath] -> SeenTree -> [([FilePath], FileId)]
subFiles xs st = case findNode xs st of
  Nothing -> []
  Just n -> allFiles n
