{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Test (
  runTest
  ) where

import F4L
import F4L.Actions
import F4L.BufferMonad
import F4L.FSMonad hiding( Error )
import F4L.Knowledge
import F4L.Interactive
import F4L.MockFSMonad
import F4L.Op
import qualified F4L.SeenTree as ST
import F4L.Utils

import Control.Concurrent( threadDelay )
import Control.DeepSeq( deepseq, force, rnf )
-- import Control.Monad
import Control.Monad.Error
import Control.Monad.State.Strict
import Control.Lens
import qualified Data.Binary as B
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BL
import Data.List( foldl', intersperse, (\\) )
import Data.Maybe( catMaybes, isNothing, fromJust )
import qualified Data.Map.Strict as SM
import Data.Set( Set )
import qualified Data.Set as S
-- import qualified Data.Trie as T
import System.Console.Haskeline( replacement )
import System.IO hiding( hShow )
-- import Text.Parsec

fid i = FileId i $ sha1 $ s2bl $ show i

s2bs s = BS.pack $ map ( toEnum . fromEnum ) s
s2bl s = BL.pack $ map ( toEnum . fromEnum ) s

data TTest a = Fast { shortcut :: String, ttest :: a } | Slow { shortcut :: String, ttest :: a }

isFast ( Fast _ _ ) = True
isFast _ = False

tests :: [TTest (Test String)]
tests = [
    Fast "fi" testFidInsert
  , Fast "fc" testFidConcat
  , Fast "jd" testJoinDir
  , Fast "af100" $ testAddFiles 100
  , Slow "af5e4" $ testAddFiles 50000
  , Fast "af2s100" $ testAddFiles2Step 100
  , Slow "af2s5e4" $ testAddFiles2Step 50000
  , Fast "res" testRestore
  , Fast "mres-100-10" $ testMultiRestore 100 10
  , Slow "mres-1e4-10" $ testMultiRestore 10000 10
  , Slow "mres-1e5-100" $ testMultiRestore 100000 100
  , Fast "rescan" testWithRescan
  , Fast "re1" testRequirements
  , Fast "kd-1e4" $ testKnownData 10000
  , Slow "kd-1e5" $ testKnownData 100000
  , Slow "kd-3e5" $ testKnownData 300000
  , Slow "kd-1e6" $ testKnownData 1000000
  , Fast "c-cwd" testCwdCompletion
  , Fast "c-st-100" $ testStorageCompletion 100
  , Slow "c-st-1e4" $ testStorageCompletion 10000
  , Fast "st-100" $ testSeenTree 100
  , Fast "st-1e4" $ testSeenTree 10000
  , Slow "st-1e5" $ testSeenTree 100000
  , Slow "st-1e6" $ testSeenTree 1000000
  , Fast "k-ser" testKnowledgeSerialization
  ]

runTest :: [String] -> IO ()
runTest ["list"] = mapM_ putStrLn $ [( reverse $ take 2 $ reverse $ ' ' : show i ) ++ " - " ++ shortcut t ++ " - " ++ test_name ( ttest t ) | (i, t) <- zip [1..] tests ]
runTest ["all"] = runSomeTests tests
runTest [] = runSomeTests $ filter isFast tests
runTest xs = runSomeTests [ maybe ( error $ x ++ " not found" ) id $ flip lookup els x | x <- xs ]
  where
    els =  join [[(shortcut t, t), (show i, t)] | (i, t) <- zip [1..] tests ]

runSomeTests someTests = do
  res <- mapM ( run_it . ( pair shortcut ttest ) ) someTests
  putStrLn $ "Passed " ++ show ( sum $ map snd res ) ++ "/" ++ show ( length res ) ++ " tests"
  case map fst $ filter ( (== 0) . snd ) res of
    [] -> return ()
    failed -> putStrLn $ "Failed: " ++ concat ( intersperse " " failed )
  where
    pair f g a = ( f a, g a )
    run_it ( sc, Test label act ) = do
      putStrLn $ "Running test: " ++ label
      res <- runErrorT act
      case res of
        Right _ -> putStrLn "OK" >> return ( sc, 1 )
        Left msg -> putStrLn ( "ERR: " ++ msg ) >> return ( sc, 0 )


data Test err = Test { test_name :: String, test_run :: ErrorT err IO () }


terminates :: ( Error err, Show a ) => String -> IO a -> Test err
terminates label act = Test label $ do
  a <- lift act
  return $ rnf ( show a )


defaultState :: F4LState KnownData ()
defaultState = f4lState (MetaDB "/tmp/no-f4l-db") (Verbosity Chatty True) UseDefault False False ()

{-
testTrie1 :: Test String
testTrie1 = terminates "1000000 trie insertions" $ do
  return $ foldl' ( \ t a -> T.insert ( s2bs $ show a ) ( fid a ) t ) T.empty [ 1 .. 1000000 ]
  
testTrie2 = terminates "10000 trie insertions" $ do
  execStateT ( mapM step [1..10000] ) ( T.empty )
  where 
    step :: Int -> StateT ( T.Trie FileId ) IO ()
    step i = do
      modify $ T.insert ( s2bs $ show i ) ( fid i )
-}

testFidInsert = terminates "10000 fid insertions" $ do
  execStateT ( mapM step [1..10000] ) defaultState
  where
    step i = do
      let f = fid i
      show f `seq` do
        st_opEnv . openv_fid .= Just f
        st_opEnv . openv_fids %= withFid f

testFidConcat = terminates "10000 fid show" $ do
  return $ concat $ map ( show . fid ) [ 1 .. 10000 ]


testJoinDir = Test "Join dir" $ mapM_ check [
  ( "/a/b", "..", "/a" )
  , ( "a/b", "..", "a" )
  , ( "/", "..", "/" )
  ]
  where
    check (a, b, c) =
      let j = joinDir a b
      in
       unless ( j == c ) ( throwError $ a ++ " join " ++ b ++ " = " ++ j ++ " <> " ++ c )

readFromFile' = onErrorWarnWithDefault "" . readFromFile


runMock :: ( MockFSState mfs, Knowledge k, Error err )
           => Int -> ErrorT err ( TopF4LMonad k ( MockFS mfs ) ) a -> ErrorT err ( StateT mfs IO ) a
runMock delta_t act =
  let state = f4lState (MetaDB "/home/user/.f4l/db") ( Verbosity Quiet True ) UseDefault False False ()
  in ErrorT $ runMockFS $ evalStateT ( runErrorT $ ( lift $ lift $ mfs_now .= delta_t ) >> act ) state

inSession' :: ( FSMonad m, BufferMonad m, Knowledge k ) => F4LMonad k m a -> GenF4LMonad k s m a
inSession' = inSession "some_env"

vfs dir n = Overlay emptyMemFS $ VirtualFS dir n

vfs2 dir1 n1 dir2 n2 = Overlay emptyMemFS $ Overlay ( VirtualFS dir1 n1 ) ( VirtualFS dir2 n2 )

mockTest :: ( MockFSState mfs, Error err ) 
            => String -> mfs -> ErrorT err ( StateT mfs IO ) a -> Test err
mockTest label fs act = Test label $ void $ ErrorT $ evalStateT ( runErrorT act ) fs


data SeenFiles = SeenFiles {
    sf_seen :: !( Set FilePath ) }

instance Knowledge SeenFiles where
  noKnowledge = SeenFiles S.empty

  applyFact r @ ( FileSeen _ _ _ _ ) s =
    s { sf_seen = S.insert (fact_path r) $ sf_seen s }
  applyFact _ s = s


interactionK :: InteractionKnowledge
interactionK = undefined

knownK :: KnownData
knownK = undefined

seenStoragesK :: SeenStorages
seenStoragesK = undefined

statsK :: ( SeenFiles, FactCounter )
statsK = undefined


testAddFiles n = mockTest ( "Adding " ++ show n ++ " files" ) ( vfs2 "/some/dir" n "/some/other" n ) ( runMock 0 act1 >> runMock 10 act2 )
  where
    act1 = do
      lift $ do
        withKnowledge interactionK
        st_interactive .= Use "2"
        inSession' $ addFiles Nothing ["/some/dir", "/some/other"]
      unlessM ( lift $ doesFileExist "/some/.f4l-id" ) $ throwError "No id file was created"
      unlessM ( lift $ doesFileExist "/home/user/.f4l/db" ) $ throwError "No id file was created"
    act2 = do
      lift $ withKnowledge statsK
      lift $ onErrorDie $ fromDB loadOps
      i <- lift $ use $ st_knowledge . fact_counter
      SeenFiles seen_paths <- lift $ use $ st_knowledge . _1
      flip mapM_ ( liftM2 (++) ["dir/", "other/"] $ map show [1..n] ) $ \ f -> do
        unless ( f `S.member` seen_paths ) $ throwError $ f ++ " not found in facts"

testAddFiles2Step n = mockTest ( "Adding " ++ show n ++ " files - 2 step" ) ( vfs2 "/some/dir" n "/some/other" n ) ( runMock 0 act1 >> runMock 10 act2 >> runMock 20 act3 >> runMock 30 act4 )
  where
    act1 = do
      lift $ do
        withKnowledge interactionK
        st_interactive .= Use "2"
        inSession' $ addFiles Nothing ["/some/dir"]
      unlessM ( lift $ doesFileExist "/some/.f4l-id" ) $ throwError "No id file was created"
      unlessM ( lift $ doesFileExist "/home/user/.f4l/db" ) $ throwError "No id file was created"
    act2 = do
      lift $ onErrorDie $ fromDB loadOps
      FactCounter i <- lift $ use st_knowledge
      unless (i > n) $ throwError "Facts not found in DB"
    act3 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/other"]
    act4 = do
      lift $ onErrorDie $ fromDB loadOps
      SeenFiles seen_paths <- lift $ use st_knowledge
      flip mapM_ ( liftM2 (++) ["dir/", "other/"] $ map show [1..n] ) $ \ f -> do
        unless ( f `S.member` seen_paths ) $ throwError $ f ++ " not found in facts"

testRestore = mockTest "Adding and restoring files" ( vfs "/some/dir" 100 ) ( runMock 0 act1 >> runMock 100 act2 >> runMock 200 act3 >> runMock 300 act4 )
  where
    act1 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/dir"]
    act2 = do
      lift $ do
        withKnowledge interactionK
        sid <- readFromFile' "/some/dir/.f4l-id"
        onErrorDie $ fromDB loadOps
        inSession' $ onErrorWarn $ restore True sid "" "/some/dir_copy" >> return ()
      forM_ (map show [1..100]) $ \ f -> do
        orig <- lift $ readFromFile' $ "/some/dir/" ++ f
        copy <- lift $ readFromFile' $ "/some/dir_copy/" ++ f
        when ( orig /= copy ) $ throwError $ "File " ++ f ++ " not restored"
    act3 = lift $ do
      withKnowledge interactionK
      sequence_ [ onErrorWarn $ removeFile $ "/some/dir/" ++ show i | i <- [10, 20..100] ]
      onErrorDie $ fromDB loadOps
      inSession' $ addFiles Nothing ["/some/dir"]
    act4 = do
      lift $ do
        withKnowledge interactionK
        sid <- readFromFile' "/some/dir/.f4l-id"
        onErrorDie $ fromDB loadOps
        inSession' $ onErrorWarn $ restore True sid "" "/some/dir_copy2" >> return ()
      forM_ (map show $ [1..100] \\ [10, 20 .. 100]) $ \ f -> do
        orig <- lift $ readFromFile' $ "/some/dir/" ++ f
        copy <- lift $ readFromFile' $ "/some/dir_copy2/" ++ f
        when ( orig /= copy ) $ throwError $ "File " ++ f ++ " not restored"
      forM_ [ "/some/dir_copy2/" ++ show i | i <- [10, 20 .. 100] ] $ \ f -> do
        whenM ( lift $ doesFileExist f ) $ throwError $ "File " ++ f ++ " should not have been restored"
      

testMultiRestore n t = mockTest ( "Restoring " ++ show n ++ " files " ++ show t ++ " times" ) ( vfs "/some/dir" n ) ( runMock 0 act1 >>= \ sid -> flip mapM_ [1..t] (\ i -> runMock (100 + i * 10) (act2 sid i)) )
  where 
    act1 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/dir"]
      readFromFile' "/some/dir/.f4l-id"
    act2 sid i = do
      lift $ do
        withKnowledge interactionK
        outputLineMsg $ "Restoring to " ++ show i
        inSession' $ onErrorWarn $ restore True sid "" ( "/some/dir_" ++ show i ) >> return ()
        cleanLine


testWithRescan = mockTest "Reindexing files with rescan" ( vfs "/some/dir" 100 ) $ runMock 0 act1 >>= \ fid -> runMock 10 ( act2 fid ) >> runMock 20 ( act3 fid )
  where
    mutableFid :: Monad m => GenF4LMonad InteractionKnowledge s m FileId
    mutableFid = use $ st_knowledge . kd_contents . to ( head . SM.elems ) . to ( fromJust . ST.findFileNode "mutable" )
    act1 = lift $ do
      withKnowledge interactionK
      onErrorWarn $ appendToFile "/some/dir/mutable" "abc"
      inSession' $ addFiles Nothing ["/some/dir"]
      mutableFid
    act2 fid = do
      lift $ withKnowledge interactionK
      lift $ onErrorWarn $ do
        removeFile "/some/dir/mutable"
        appendToFile "/some/dir/mutable" "ABC"
      lift $ do
        onErrorDie $ fromDB loadOps        
        inSession' $ addFiles Nothing ["/some/dir"]
      n_fid <- lift mutableFid
      when (n_fid /= fid) $ throwError "Expecting file change NOT to be detected"
    act3 fid = do
      lift $ do
        onErrorDie $ fromDB loadOps
        st_withRescan .= True
        inSession' $ addFiles Nothing ["/some/dir"]
      n_fid <- lift mutableFid
      when (n_fid == fid) $ throwError "Expecting file change to be detected with rescan"


testRequirements = mockTest ( "Setting and changing requirements" ) ( vfs "/some/dir" 10 ) ( runMock 0 act1 >> runMock 10 act2 )
  where 
    act1 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/dir"]
    act2 = do
      lift $ withKnowledge interactionK
      f <- lift $ use $ st_knowledge . kd_files
      when ( any ( ( /= DefaultPriority ) . fd_priority ) $ SM.elems f ) $ throwError "Added files don't have default priority"
      lift $ inSession' $ addFiles ( Just AsManyAsPossible ) ["/some/dir"]
      f <- lift $ use $ st_knowledge . kd_files
      when ( any ( ( /= AsManyAsPossible ) . fd_priority ) $ SM.elems f ) $ throwError "Re-added files don't have max priority"


testKnownData n = terminates ( "Generating known data for " ++ show n ++ " files" ) act
  where
    act = print $ length $ show $ bigData
    boh :: KnownData
    boh = noKnowledge
    bigData = foldl' ( flip applyFact ) boh lotsOfFacts
    lotsOfFacts = flip map [1..n] $ \ i -> 
      FileSeen 0 ( FileId i $ sha1 $ BL.pack $ map (toEnum . fromEnum) $ show i ) ( take 10 $ show i ++ "abcdefghi" ) ( "/some/path/" ++ show i )


testCwdCompletion = mockTest "Command completion on local dir" ( vfs "/some/dir" 100 ) ( runMock 0 act )
  where
    comp s = flip evalStateT s $ iComplete (reverse "ls /some/dir/", "")
    act = do
      iState <- lift initialInteractionStatus
      (_, completions) <- lift $ inSession' $ comp iState
      let rs = map replacement completions
      case filter (not . (`elem` rs)) ["/some/dir/" ++ show i | i <- [1..100]] of
        [] -> return ()
        x -> throwError $ show ( length x ) ++ "/100 files not present in completion"


testStorageCompletion n = mockTest ( "Command completion on storage with " ++ show n ++ " files" ) ( vfs "/some/dir" n ) ( runMock 0 act1 >>= runMock 10 . act2 )
  where
    comp s = flip evalStateT s $ iComplete (reverse "ls /", "")
    act1 = do
      lift $ withKnowledge interactionK
      lift $ inSession' $ addFiles Nothing ["/some/dir"]
      esid <- lift $ readFromFile' "/some/dir/.f4l-id" 
      if esid == "" then ( throwError "Sid not found in dir" ) else return esid
    act2 sid = do
      lift $ onErrorDie $ fromDB loadOps
      iState <- lift initialInteractionStatus
      (_, completions) <- lift $ inSession' $ comp $ set is_context ( InStorage sid ) iState
      let rs = map replacement completions
      case filter (not . (`elem` rs)) ["/" ++ show i | i <- [1..n]] of
        [] -> return ()
        x -> throwError $ show ( length x ) ++ "/" ++ show n ++ " files not present in completion"


testSeenTree :: Int -> Test String
testSeenTree n = terminates ( "Generating SeenTree with " ++ show n ++ " files" ) $ do
  let st = force ( foldl' ins ST.empty $ zip [1..] $ map testFid [1..n] )
  putStrLn "Computing length"
  putStrLn $ hShow $ fromIntegral $ BL.length ( B.encode st )
  where
    ins st (i, a) = ST.insertFile (show i) a st


testKnowledgeSerialization :: Test String
testKnowledgeSerialization = mockTest ( "Knowledge serialization" ) ( vfs "/some/dir" 1000 ) ( runMock 0 act1 )
  where
    act1 = do
      lift $ withKnowledge (seenStoragesK, knownK)
      lift $ inSession' ( addFiles Nothing ["/some/dir"] ) >> st_opEnv .= initialOpEnv >> st_knowledge .= noKnowledge >> onErrorDie saveCache
      lift $ onErrorDie loadCached
